.App{text-align:left}.navBar{background-color:#fc0}.btnGrey{background-color:#bcbcbb}.btnBlue{background-color:#00a6e1}.btnGreen{background-color:#5dd41b}.btnRed{background-color:#df462e}.fontSizeSmall{font-size:90%!important}.fontSizeDefault{font-size:100%!important}.fontSizeLarge{font-size:120%!important}.fontSizeLargest{font-size:140%!important}.tooltip-card{transition:box-shadow .3s cubic-bezier(.4,0,.2,1) 0ms;border-radius:4px;overflow:hidden}.tooltip-card-content{padding:16px 16px 24px}.tooltip-title{font-family:"Roboto","Helvetica","Arial",sans-serif;font-weight:500;font-size:.875rem;line-height:1.57;letter-spacing:.00714em;margin:0 0 16px}.tooltip-label-marker{height:8px;width:8px;border-width:0;border-color:#fff;border-radius:12px}.tooltip-label-text{left:4px;font-size:14px;font-weight:500;font-family:Roboto,Helvetica,Arial,sans-serif}.tooltip-label-number{font-weight:500;font-size:1.25rem;line-height:1.6;letter-spacing:.0075em}.tooltip-label-number,.tooltip-label-text-secondary{margin:0;font-family:"Roboto","Helvetica","Arial",sans-serif}.tooltip-label-text-secondary{font-weight:400;font-size:.875rem;line-height:1.43;letter-spacing:.01071em;max-width:240px;word-wrap:break-word;white-space:normal}.tooltip-divider{flex-shrink:0;border-style:solid;border-width:0 0 thin;margin:16px 0}.tooltip-img{width:90px;height:60px;margin:6px}.apexcharts-tooltip{box-shadow:0 2px 1px -1px rgb(0 0 0/20%),0 1px 1px 0 rgb(0 0 0/14%),0 1px 3px 0 rgb(0 0 0/12%)}.MuiBanner-cardContentIconAppBar{padding-top:8px!important}.rotate{animation:rotation 8s linear infinite}@keyframes rotation{0%{transform:rotate(0deg)}to{transform:rotate(359deg)}}
/*# sourceMappingURL=main.ac60f49f.chunk.css.map */